@@ -35,7 +35,7 @@ from wsgiref.simple_server import (
     ServerHandler,
     WSGIServer,
     make_server,
-    )
+)
 
 from urllib.parse import parse_qs
 
@@ -43,28 +43,28 @@ from urllib.parse import parse_qs
 from dulwich import log_utils
 from dulwich.protocol import (
     ReceivableProtocol,
-    )
+)
 from dulwich.repo import (
     BaseRepo,
     NotGitRepository,
     Repo,
-    )
+)
 from dulwich.server import (
     DictBackend,
     DEFAULT_HANDLERS,
     generate_info_refs,
     generate_objects_info_packs,
-    )
+)
 
 
 logger = log_utils.getLogger(__name__)
 
 
 # HTTP error strings
-HTTP_OK = '200 OK'
-HTTP_NOT_FOUND = '404 Not Found'
-HTTP_FORBIDDEN = '403 Forbidden'
-HTTP_ERROR = '500 Internal Server Error'
+HTTP_OK = "200 OK"
+HTTP_NOT_FOUND = "404 Not Found"
+HTTP_FORBIDDEN = "403 Forbidden"
+HTTP_ERROR = "500 Internal Server Error"
 
 
 def date_time_string(timestamp: Optional[float] = None) -> str:
@@ -74,15 +74,34 @@ def date_time_string(timestamp: Optional[float] = None) -> str:
     #  - weekdayname and monthname are renamed and locals rather than class
     #    variables.
     # Copyright (c) 2001-2010 Python Software Foundation; All Rights Reserved
-    weekdays = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']
-    months = [None,
-              'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
-              'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']
+    weekdays = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]
+    months = [
+        None,
+        "Jan",
+        "Feb",
+        "Mar",
+        "Apr",
+        "May",
+        "Jun",
+        "Jul",
+        "Aug",
+        "Sep",
+        "Oct",
+        "Nov",
+        "Dec",
+    ]
     if timestamp is None:
         timestamp = time.time()
     year, month, day, hh, mm, ss, wd = time.gmtime(timestamp)[:7]
-    return '%s, %02d %3s %4d %02d:%02d:%02d GMD' % (
-            weekdays[wd], day, months[month], year, hh, mm, ss)
+    return "%s, %02d %3s %4d %02d:%02d:%02d GMD" % (
+        weekdays[wd],
+        day,
+        months[month],
+        year,
+        hh,
+        mm,
+        ss,
+    )
 
 
 def url_prefix(mat) -> str:
@@ -94,7 +113,7 @@ def url_prefix(mat) -> str:
         original string. Normalized to start with one leading slash and end
         with zero.
     """
-    return '/' + mat.string[:mat.start()].strip('/')
+    return "/" + mat.string[: mat.start()].strip("/")
 
 
 def get_repo(backend, mat) -> BaseRepo:
@@ -112,7 +131,7 @@ def send_file(req, f, content_type):
     Returns: Iterator over the contents of the file, as chunks.
     """
     if f is None:
-        yield req.not_found('File not found')
+        yield req.not_found("File not found")
         return
     try:
         req.respond(HTTP_OK, content_type)
@@ -122,96 +141,132 @@ def send_file(req, f, content_type):
                 break
             yield data
     except IOError:
-        yield req.error('Error reading file')
+        yield req.error("Error reading file")
     finally:
         f.close()
 
 
 def _url_to_path(url):
-    return url.replace('/', os.path.sep)
+    return url.replace("/", os.path.sep)
 
 
 def get_text_file(req, backend, mat):
     req.nocache()
     path = _url_to_path(mat.group())
-    logger.info('Sending plain text file %s', path)
-    return send_file(req, get_repo(backend, mat).get_named_file(path),
-                     'text/plain')
+    logger.info("Sending plain text file %s", path)
+    return send_file(req, get_repo(backend, mat).get_named_file(path), "text/plain")
 
 
 def get_loose_object(req, backend, mat):
-    sha = (mat.group(1) + mat.group(2)).encode('ascii')
-    logger.info('Sending loose object %s', sha)
+    sha = (mat.group(1) + mat.group(2)).encode("ascii")
+    logger.info("Sending loose object %s", sha)
     object_store = get_repo(backend, mat).object_store
     if not object_store.contains_loose(sha):
-        yield req.not_found('Object not found')
+        yield req.not_found("Object not found")
         return
     try:
         data = object_store[sha].as_legacy_object()
     except IOError:
-        yield req.error('Error reading object')
+        yield req.error("Error reading object")
         return
     req.cache_forever()
-    req.respond(HTTP_OK, 'application/x-git-loose-object')
+    req.respond(HTTP_OK, "application/x-git-loose-object")
     yield data
 
 
 def get_pack_file(req, backend, mat):
     req.cache_forever()
     path = _url_to_path(mat.group())
-    logger.info('Sending pack file %s', path)
-    return send_file(req, get_repo(backend, mat).get_named_file(path),
-                     'application/x-git-packed-objects')
+    logger.info("Sending pack file %s", path)
+    return send_file(
+        req,
+        get_repo(backend, mat).get_named_file(path),
+        "application/x-git-packed-objects",
+    )
 
 
 def get_idx_file(req, backend, mat):
     req.cache_forever()
     path = _url_to_path(mat.group())
-    logger.info('Sending pack file %s', path)
-    return send_file(req, get_repo(backend, mat).get_named_file(path),
-                     'application/x-git-packed-objects-toc')
+    logger.info("Sending pack file %s", path)
+    return send_file(
+        req,
+        get_repo(backend, mat).get_named_file(path),
+        "application/x-git-packed-objects-toc",
+    )
 
 
 def get_info_refs(req, backend, mat):
-    params = parse_qs(req.environ['QUERY_STRING'])
-    service = params.get('service', [None])[0]
+    params = parse_qs(req.environ["QUERY_STRING"])
+    service = params.get("service", [None])[0]
     try:
         repo = get_repo(backend, mat)
     except NotGitRepository as e:
         yield req.not_found(str(e))
         return
     if service and not req.dumb:
-        handler_cls = req.handlers.get(service.encode('ascii'), None)
+        handler_cls = req.handlers.get(service.encode("ascii"), None)
         if handler_cls is None:
-            yield req.forbidden('Unsupported service')
+            yield req.forbidden("Unsupported service")
             return
         req.nocache()
-        write = req.respond(
-            HTTP_OK, 'application/x-%s-advertisement' % service)
+        write = req.respond(HTTP_OK, "application/x-%s-advertisement" % service)
         proto = ReceivableProtocol(BytesIO().read, write)
-        handler = handler_cls(backend, [url_prefix(mat)], proto,
-                              http_req=req, advertise_refs=True)
-        handler.proto.write_pkt_line(
-            b'# service=' + service.encode('ascii') + b'\n')
+        handler = handler_cls(
+            backend,
+            [url_prefix(mat)],
+            proto,
+            stateless_rpc=req,
+            advertise_refs=True,
+        )
+        handler.proto.write_pkt_line(b"# service=" + service.encode("ascii") + b"\n")
         handler.proto.write_pkt_line(None)
         handler.handle()
     else:
         # non-smart fallback
         # TODO: select_getanyfile() (see http-backend.c)
         req.nocache()
-        req.respond(HTTP_OK, 'text/plain')
-        logger.info('Emulating dumb info/refs')
+        req.respond(HTTP_OK, "text/plain")
+        logger.info("Emulating dumb info/refs")
         for text in generate_info_refs(repo):
             yield text
 
 
 def get_info_packs(req, backend, mat):
     req.nocache()
-    req.respond(HTTP_OK, 'text/plain')
-    logger.info('Emulating dumb info/packs')
+    req.respond(HTTP_OK, "text/plain")
+    logger.info("Emulating dumb info/packs")
     return generate_objects_info_packs(get_repo(backend, mat))
 
 
+def _chunk_iter(f):
+    while True:
+        line = f.readline()
+        length = int(line.rstrip(), 16)
+        chunk = f.read(length + 2)
+        if length == 0:
+            break
+        yield chunk[:-2]
+
+
+class ChunkReader(object):
+
+    def __init__(self, f):
+        self._iter = _chunk_iter(f)
+        self._buffer = []
+
+    def read(self, n):
+        while sum(map(len, self._buffer)) < n:
+            try:
+                self._buffer.append(next(self._iter))
+            except StopIteration:
+                break
+        f = b''.join(self._buffer)
+        ret = f[:n]
+        self._buffer = [f[n:]]
+        return ret
+
+
 class _LengthLimitedFile(object):
     """Wrapper class to limit the length of reads from a file-like object.
 
@@ -226,7 +281,7 @@ class _LengthLimitedFile(object):
 
     def read(self, size=-1):
         if self._bytes_avail <= 0:
-            return b''
+            return b""
         if size == -1 or size > self._bytes_avail:
             size = self._bytes_avail
         self._bytes_avail -= size
@@ -236,11 +291,11 @@ class _LengthLimitedFile(object):
 
 
 def handle_service_request(req, backend, mat):
-    service = mat.group().lstrip('/')
-    logger.info('Handling service request for %s', service)
-    handler_cls = req.handlers.get(service.encode('ascii'), None)
+    service = mat.group().lstrip("/")
+    logger.info("Handling service request for %s", service)
+    handler_cls = req.handlers.get(service.encode("ascii"), None)
     if handler_cls is None:
-        yield req.forbidden('Unsupported service')
+        yield req.forbidden("Unsupported service")
         return
     try:
         get_repo(backend, mat)
@@ -248,22 +303,26 @@ def handle_service_request(req, backend, mat):
         yield req.not_found(str(e))
         return
     req.nocache()
-    write = req.respond(HTTP_OK, 'application/x-%s-result' % service)
-    proto = ReceivableProtocol(req.environ['wsgi.input'].read, write)
+    write = req.respond(HTTP_OK, "application/x-%s-result" % service)
+    if req.environ.get('HTTP_TRANSFER_ENCODING') == 'chunked':
+        read = ChunkReader(req.environ["wsgi.input"]).read
+    else:
+        read = req.environ["wsgi.input"].read
+    proto = ReceivableProtocol(read, write)
     # TODO(jelmer): Find a way to pass in repo, rather than having handler_cls
     # reopen.
-    handler = handler_cls(backend, [url_prefix(mat)], proto, http_req=req)
+    handler = handler_cls(backend, [url_prefix(mat)], proto, stateless_rpc=req)
     handler.handle()
 
 
 class HTTPGitRequest(object):
     """Class encapsulating the state of a single git HTTP request.
 
-    :ivar environ: the WSGI environment for the request.
+    Attributes:
+      environ: the WSGI environment for the request.
     """
 
-    def __init__(
-            self, environ, start_response, dumb: bool = False, handlers=None):
+    def __init__(self, environ, start_response, dumb: bool = False, handlers=None):
         self.environ = environ
         self.dumb = dumb
         self.handlers = handlers
@@ -276,14 +335,16 @@ class HTTPGitRequest(object):
         self._headers.append((name, value))
 
     def respond(
-            self, status: str = HTTP_OK,
-            content_type: Optional[str] = None,
-            headers: Optional[List[Tuple[str, str]]] = None):
+        self,
+        status: str = HTTP_OK,
+        content_type: Optional[str] = None,
+        headers: Optional[List[Tuple[str, str]]] = None,
+    ):
         """Begin a response with the given status and other headers."""
         if headers:
             self._headers.extend(headers)
         if content_type:
-            self._headers.append(('Content-Type', content_type))
+            self._headers.append(("Content-Type", content_type))
         self._headers.extend(self._cache_headers)
 
         return self._start_response(status, self._headers)
@@ -291,68 +352,72 @@ class HTTPGitRequest(object):
     def not_found(self, message: str) -> bytes:
         """Begin a HTTP 404 response and return the text of a message."""
         self._cache_headers = []
-        logger.info('Not found: %s', message)
-        self.respond(HTTP_NOT_FOUND, 'text/plain')
-        return message.encode('ascii')
+        logger.info("Not found: %s", message)
+        self.respond(HTTP_NOT_FOUND, "text/plain")
+        return message.encode("ascii")
 
     def forbidden(self, message: str) -> bytes:
         """Begin a HTTP 403 response and return the text of a message."""
         self._cache_headers = []
-        logger.info('Forbidden: %s', message)
-        self.respond(HTTP_FORBIDDEN, 'text/plain')
-        return message.encode('ascii')
+        logger.info("Forbidden: %s", message)
+        self.respond(HTTP_FORBIDDEN, "text/plain")
+        return message.encode("ascii")
 
     def error(self, message: str) -> bytes:
         """Begin a HTTP 500 response and return the text of a message."""
         self._cache_headers = []
-        logger.error('Error: %s', message)
-        self.respond(HTTP_ERROR, 'text/plain')
-        return message.encode('ascii')
+        logger.error("Error: %s", message)
+        self.respond(HTTP_ERROR, "text/plain")
+        return message.encode("ascii")
 
     def nocache(self) -> None:
         """Set the response to never be cached by the client."""
         self._cache_headers = [
-          ('Expires', 'Fri, 01 Jan 1980 00:00:00 GMT'),
-          ('Pragma', 'no-cache'),
-          ('Cache-Control', 'no-cache, max-age=0, must-revalidate'),
-          ]
+            ("Expires", "Fri, 01 Jan 1980 00:00:00 GMT"),
+            ("Pragma", "no-cache"),
+            ("Cache-Control", "no-cache, max-age=0, must-revalidate"),
+        ]
 
     def cache_forever(self) -> None:
         """Set the response to be cached forever by the client."""
         now = time.time()
         self._cache_headers = [
-          ('Date', date_time_string(now)),
-          ('Expires', date_time_string(now + 31536000)),
-          ('Cache-Control', 'public, max-age=31536000'),
-          ]
+            ("Date", date_time_string(now)),
+            ("Expires", date_time_string(now + 31536000)),
+            ("Cache-Control", "public, max-age=31536000"),
+        ]
 
 
 class HTTPGitApplication(object):
     """Class encapsulating the state of a git WSGI application.
 
-    :ivar backend: the Backend object backing this application
+    Attributes:
+      backend: the Backend object backing this application
     """
 
     services = {
-      ('GET', re.compile('/HEAD$')): get_text_file,
-      ('GET', re.compile('/info/refs$')): get_info_refs,
-      ('GET', re.compile('/objects/info/alternates$')): get_text_file,
-      ('GET', re.compile('/objects/info/http-alternates$')): get_text_file,
-      ('GET', re.compile('/objects/info/packs$')): get_info_packs,
-      ('GET', re.compile('/objects/([0-9a-f]{2})/([0-9a-f]{38})$')):
-      get_loose_object,
-      ('GET', re.compile('/objects/pack/pack-([0-9a-f]{40})\\.pack$')):
-      get_pack_file,
-      ('GET', re.compile('/objects/pack/pack-([0-9a-f]{40})\\.idx$')):
-      get_idx_file,
-
-      ('POST', re.compile('/git-upload-pack$')): handle_service_request,
-      ('POST', re.compile('/git-receive-pack$')): handle_service_request,
+        ("GET", re.compile("/HEAD$")): get_text_file,
+        ("GET", re.compile("/info/refs$")): get_info_refs,
+        ("GET", re.compile("/objects/info/alternates$")): get_text_file,
+        ("GET", re.compile("/objects/info/http-alternates$")): get_text_file,
+        ("GET", re.compile("/objects/info/packs$")): get_info_packs,
+        (
+            "GET",
+            re.compile("/objects/([0-9a-f]{2})/([0-9a-f]{38})$"),
+        ): get_loose_object,
+        (
+            "GET",
+            re.compile("/objects/pack/pack-([0-9a-f]{40})\\.pack$"),
+        ): get_pack_file,
+        (
+            "GET",
+            re.compile("/objects/pack/pack-([0-9a-f]{40})\\.idx$"),
+        ): get_idx_file,
+        ("POST", re.compile("/git-upload-pack$")): handle_service_request,
+        ("POST", re.compile("/git-receive-pack$")): handle_service_request,
     }
 
-    def __init__(
-            self, backend, dumb: bool = False, handlers=None,
-            fallback_app=None):
+    def __init__(self, backend, dumb: bool = False, handlers=None, fallback_app=None):
         self.backend = backend
         self.dumb = dumb
         self.handlers = dict(DEFAULT_HANDLERS)
@@ -361,10 +426,11 @@ class HTTPGitApplication(object):
             self.handlers.update(handlers)
 
     def __call__(self, environ, start_response):
-        path = environ['PATH_INFO']
-        method = environ['REQUEST_METHOD']
-        req = HTTPGitRequest(environ, start_response, dumb=self.dumb,
-                             handlers=self.handlers)
+        path = environ["PATH_INFO"]
+        method = environ["REQUEST_METHOD"]
+        req = HTTPGitRequest(
+            environ, start_response, dumb=self.dumb, handlers=self.handlers
+        )
         # environ['QUERY_STRING'] has qs args
         handler = None
         for smethod, spath in self.services.keys():
@@ -379,7 +445,7 @@ class HTTPGitApplication(object):
             if self.fallback_app is not None:
                 return self.fallback_app(environ, start_response)
             else:
-                return [req.not_found('Sorry, that method is not supported')]
+                return [req.not_found("Sorry, that method is not supported")]
 
         return handler(req, self.backend, mat)
 
@@ -393,24 +459,25 @@ class GunzipFilter(object):
         self.app = application
 
     def __call__(self, environ, start_response):
-        if environ.get('HTTP_CONTENT_ENCODING', '') == 'gzip':
+        if environ.get("HTTP_CONTENT_ENCODING", "") == "gzip":
             try:
-                environ['wsgi.input'].tell()
-                wsgi_input = environ['wsgi.input']
+                environ["wsgi.input"].tell()
+                wsgi_input = environ["wsgi.input"]
             except (AttributeError, IOError, NotImplementedError):
                 # The gzip implementation in the standard library of Python 2.x
                 # requires working '.seek()' and '.tell()' methods on the input
                 # stream.  Read the data into a temporary file to work around
                 # this limitation.
                 wsgi_input = tempfile.SpooledTemporaryFile(16 * 1024 * 1024)
-                shutil.copyfileobj(environ['wsgi.input'], wsgi_input)
+                shutil.copyfileobj(environ["wsgi.input"], wsgi_input)
                 wsgi_input.seek(0)
 
-            environ['wsgi.input'] = gzip.GzipFile(
-                filename=None, fileobj=wsgi_input, mode='r')
-            del environ['HTTP_CONTENT_ENCODING']
-            if 'CONTENT_LENGTH' in environ:
-                del environ['CONTENT_LENGTH']
+            environ["wsgi.input"] = gzip.GzipFile(
+                filename=None, fileobj=wsgi_input, mode="r"
+            )
+            del environ["HTTP_CONTENT_ENCODING"]
+            if "CONTENT_LENGTH" in environ:
+                del environ["CONTENT_LENGTH"]
 
         return self.app(environ, start_response)
 
@@ -428,10 +495,11 @@ class LimitedInputFilter(object):
         # server. Unfortunately, there's no way to tell that at this point.
         # TODO: git may used HTTP/1.1 chunked encoding instead of specifying
         # content-length
-        content_length = environ.get('CONTENT_LENGTH', '')
+        content_length = environ.get("CONTENT_LENGTH", "")
         if content_length:
-            environ['wsgi.input'] = _LengthLimitedFile(
-                environ['wsgi.input'], int(content_length))
+            environ["wsgi.input"] = _LengthLimitedFile(
+                environ["wsgi.input"], int(content_length)
+            )
         return self.app(environ, start_response)
 
 
@@ -448,8 +516,10 @@ class ServerHandlerLogger(ServerHandler):
     """ServerHandler that uses dulwich's logger for logging exceptions."""
 
     def log_exception(self, exc_info):
-        logger.exception('Exception happened during processing of request',
-                         exc_info=exc_info)
+        logger.exception(
+            "Exception happened during processing of request",
+            exc_info=exc_info,
+        )
 
     def log_message(self, format, *args):
         logger.info(format, *args)
@@ -462,8 +532,10 @@ class WSGIRequestHandlerLogger(WSGIRequestHandler):
     """WSGIRequestHandler that uses dulwich's logger for logging exceptions."""
 
     def log_exception(self, exc_info):
-        logger.exception('Exception happened during processing of request',
-                         exc_info=exc_info)
+        logger.exception(
+            "Exception happened during processing of request",
+            exc_info=exc_info,
+        )
 
     def log_message(self, format, *args):
         logger.info(format, *args)
@@ -481,29 +553,39 @@ class WSGIRequestHandlerLogger(WSGIRequestHandler):
         handler = ServerHandlerLogger(
             self.rfile, self.wfile, self.get_stderr(), self.get_environ()
         )
-        handler.request_handler = self      # backpointer for logging
+        handler.request_handler = self  # backpointer for logging
         handler.run(self.server.get_app())
 
 
 class WSGIServerLogger(WSGIServer):
-
     def handle_error(self, request, client_address):
         """Handle an error. """
         logger.exception(
-            'Exception happened during processing of request from %s' %
-            str(client_address))
+            "Exception happened during processing of request from %s"
+            % str(client_address)
+        )
 
 
 def main(argv=sys.argv):
     """Entry point for starting an HTTP git server."""
     import optparse
+
     parser = optparse.OptionParser()
-    parser.add_option("-l", "--listen_address", dest="listen_address",
-                      default="localhost",
-                      help="Binding IP address.")
-    parser.add_option("-p", "--port", dest="port", type=int,
-                      default=8000,
-                      help="Port to listen on.")
+    parser.add_option(
+        "-l",
+        "--listen_address",
+        dest="listen_address",
+        default="localhost",
+        help="Binding IP address.",
+    )
+    parser.add_option(
+        "-p",
+        "--port",
+        dest="port",
+        type=int,
+        default=8000,
+        help="Port to listen on.",
+    )
     options, args = parser.parse_args(argv)
 
     if len(args) > 1:
@@ -512,15 +594,22 @@ def main(argv=sys.argv):
         gitdir = os.getcwd()
 
     log_utils.default_logging_config()
-    backend = DictBackend({'/': Repo(gitdir)})
+    backend = DictBackend({"/": Repo(gitdir)})
     app = make_wsgi_chain(backend)
-    server = make_server(options.listen_address, options.port, app,
-                         handler_class=WSGIRequestHandlerLogger,
-                         server_class=WSGIServerLogger)
-    logger.info('Listening for HTTP connections on %s:%d',
-                options.listen_address, options.port)
+    server = make_server(
+        options.listen_address,
+        options.port,
+        app,
+        handler_class=WSGIRequestHandlerLogger,
+        server_class=WSGIServerLogger,
+    )
+    logger.info(
+        "Listening for HTTP connections on %s:%d",
+        options.listen_address,
+        options.port,
+    )
     server.serve_forever()
 
 
-if __name__ == '__main__':
+if __name__ == "__main__":
     main()
